/*
 * COLOR.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements string parsing and rule application to select a
 * given set of color attributes to render any particular file with.
 *
 * Copyright (c) 2014-2017 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  This is a little circular.  Default contains both a color and an
//  instruction to use the previous window color.  If we're able to get
//  a window color, it'll pick that, but if not, the previous window color
//  is populated from the color here.
//

SDIR_COLOR_ATTRIBUTES SdirDefaultColor = {SDIR_ATTRCTRL_WINDOW_BG | SDIR_ATTRCTRL_WINDOW_FG, SDIR_DEFAULT_COLOR};

SDIR_ATTRIBUTE_COLOR_STRING ColorString[] = {

    //
    //  Some CGA color constants
    //

    {_T("black"),        {0, 0x00}},
    {_T("blue"),         {0, 0x01}},
    {_T("green"),        {0, 0x02}},
    {_T("cyan"),         {0, 0x03}},
    {_T("red"),          {0, 0x04}},
    {_T("magenta"),      {0, 0x05}},
    {_T("brown"),        {0, 0x06}},
    {_T("gray"),         {0, 0x07}},

    {_T("darkgray"),     {0, 0x08}},
    {_T("lightblue"),    {0, 0x09}},
    {_T("lightgreen"),   {0, 0x0A}},
    {_T("lightcyan"),    {0, 0x0B}},
    {_T("lightred"),     {0, 0x0C}},
    {_T("lightmagenta"), {0, 0x0D}},
    {_T("yellow"),       {0, 0x0E}},
    {_T("white"),        {0, 0x0F}},

    //
    //  Helper modifier
    //

    {_T("bright"),       {0, 0x08}},

    //
    //  Some non-color things
    //

    {_T("invert"),       {SDIR_ATTRCTRL_INVERT, 0}},
    {_T("hide"),         {SDIR_ATTRCTRL_HIDE, 0}},
    {_T("continue"),     {SDIR_ATTRCTRL_CONTINUE, 0}},
    {_T("file"),         {SDIR_ATTRCTRL_FILE, 0}},
    {_T("window_bg"),    {SDIR_ATTRCTRL_WINDOW_BG, 0}},
    {_T("window_fg"),    {SDIR_ATTRCTRL_WINDOW_FG, 0}}
};


SDIR_COLOR_ATTRIBUTES
SdirAttributeFromString(
    __in LPTSTR String
    )
{
    SDIR_COLOR_ATTRIBUTES Attribute = {0};
    DWORD Element;
    BOOL Background = FALSE;
    BOOL ExplicitBackground = FALSE;
    BOOL ExplicitForeground = FALSE;
    LPTSTR Next;

    //
    //  If it's a number, parse it like we used to.
    //

    if (String[0] >= '0' && String[0] <= '9') {
        SDIR_COLOR_ATTRIBUTES ReturnValue;
        WORD NumericRepresentation = (WORD)SdirStringToNum32(String, NULL);
        ReturnValue.Ctrl = NumericRepresentation >> 8;
        ReturnValue.Win32Attr = NumericRepresentation & 0xFF;
        return ReturnValue;
    }

    //
    //  Loop through the list of color keywords
    //

    while (String && String[0] != '\0') {

        //
        //  Check if we're setting a background.
        //
    
        Background = FALSE;
        if ((String[0] == 'b' || String[0] == 'B') &&
            (String[1] == 'g' || String[1] == 'G') &&
            String[2] == '_') {
    
            String += 3;
            Background = TRUE;
        }

        //
        //  Look for the next element
        //

        Next = _tcschr(String, '+');
        if (Next) {
            *Next = '\0';
        }

        //
        //  Walk through the string table for a match.
        //
    
        for (Element = 0; Element < sizeof(ColorString)/sizeof(ColorString[0]); Element++) {
            if (_tcsicmp(String, ColorString[Element].String) == 0) {
    
                if (Background) {
                    if (ColorString[Element].Attr.Ctrl != 0) {
                        Attribute.Ctrl |= ColorString[Element].Attr.Ctrl;
                    } else {
                        ExplicitBackground = TRUE;
                    }
                    Attribute.Win32Attr |= (ColorString[Element].Attr.Win32Attr & SDIR_ATTRIBUTE_ONECOLOR_MASK) << 4;
                } else {
                    if (ColorString[Element].Attr.Ctrl != 0) {
                        Attribute.Ctrl |= ColorString[Element].Attr.Ctrl;
                    } else {
                        ExplicitForeground = TRUE;
                    }
                    Attribute.Win32Attr |= ColorString[Element].Attr.Win32Attr;
                }
    
                break;
            }
        }
    
        //
        //  Advance to the next element
        //

        if (Next) {
            String = Next + 1;
        } else {
            String = NULL;
        }
    }

    //
    //  If an explicit background or foreground was specified,
    //  use it.  If not, assume window defaults.  When combining,
    //  if a previous color was specified explicitly, it will take
    //  precedence.
    //

    if (!ExplicitBackground) {
        Attribute.Ctrl |= SDIR_ATTRCTRL_WINDOW_BG;
    }

    if (!ExplicitForeground) {
        Attribute.Ctrl |= SDIR_ATTRCTRL_WINDOW_FG;
    }

    return Attribute;
}

BOOL
SdirColorStringFromFeature(
    __in PCSDIR_FEATURE Feature,
    __out LPTSTR String,
    __in DWORD StringSize
    )
{
    WORD Forecolor;
    WORD Backcolor;

    LPTSTR Backstring;
    LPTSTR Forestring;

    if (Feature->Flags & SDIR_FEATURE_USE_FILE_COLOR) {
        _stprintf_s(String, StringSize, _T("file"));
        return TRUE;
    }

    Forecolor = Feature->HighlightColor.Win32Attr & SDIR_ATTRIBUTE_ONECOLOR_MASK;
    Backcolor = (Feature->HighlightColor.Win32Attr >> 4) & SDIR_ATTRIBUTE_ONECOLOR_MASK;

    if (Feature->HighlightColor.Ctrl & SDIR_ATTRCTRL_WINDOW_BG) {
        Backstring = NULL;
    } else {
        Backstring = ColorString[Backcolor].String;
    }

    if (Feature->HighlightColor.Ctrl & SDIR_ATTRCTRL_WINDOW_FG) {
        Forestring = NULL;
    } else {
        Forestring = ColorString[Forecolor].String;
    }

    if (Backstring != NULL && Forestring != NULL) {
        _stprintf_s(String, StringSize, _T("bg_%s+%s"), Backstring, Forestring);
    } else if (Backstring != NULL) {
        _stprintf_s(String, StringSize, _T("bg_%s"), Backstring);
    } else if (Forestring != NULL) {
        _stprintf_s(String, StringSize, _T("%s"), Forestring);
    }

    return TRUE;
}

typedef struct _SDIR_ATTRIBUTE_APPLY {
    SDIR_COMPARE_FN CompareFn;
    BOOL TruthStates[3];
    SDIR_COLOR_ATTRIBUTES Attribute;
    WORD AlignmentPadding;
    SDIR_DIRENT CompareEntry;
} SDIR_ATTRIBUTE_APPLY, *PSDIR_ATTRIBUTE_APPLY;


BOOL
SdirPopulateApplyEntry(
    __in PSDIR_ATTRIBUTE_APPLY ThisApply,
    __in LPCTSTR Operator,
    __in LPCTSTR Criteria,
    __in DWORD FeatureNumber
    )
{
    PSDIR_FEATURE Feature;

    //
    //  Based on the operator, fill in the truth table.  We'll
    //  use the generic compare function and based on this
    //  truth table we'll decide whether the rule is satisfied
    //  or not.
    //

    if (Operator[0] == '>') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].CompareFn;

        ThisApply->TruthStates[SDIR_LESS_THAN] = FALSE;
        ThisApply->TruthStates[SDIR_GREATER_THAN] = TRUE;
        if (Operator[1] == '=') {
            ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
        } else {
            ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
        }
    } else if (Operator[0] == '<') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].CompareFn;

        ThisApply->TruthStates[SDIR_LESS_THAN] = TRUE;
        ThisApply->TruthStates[SDIR_GREATER_THAN] = FALSE;
        if (Operator[1] == '=') {
            ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
        } else {
            ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
        }
    } else if (Operator[0] == '!' && Operator[1] == '=') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].CompareFn;
        ThisApply->TruthStates[SDIR_LESS_THAN] = TRUE;
        ThisApply->TruthStates[SDIR_GREATER_THAN] = TRUE;
        ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
    } else if (Operator[0] == '=') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].CompareFn;
        ThisApply->TruthStates[SDIR_LESS_THAN] = FALSE;
        ThisApply->TruthStates[SDIR_GREATER_THAN] = FALSE;
        ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
    } else if (Operator[0] == '&') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].BitwiseCompareFn;
        ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
        ThisApply->TruthStates[SDIR_NOT_EQUAL] = FALSE;
    } else if (Operator[0] == '!' && Operator[1] == '&') {
        ThisApply->CompareFn = SdirOptions[FeatureNumber].BitwiseCompareFn;
        ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
        ThisApply->TruthStates[SDIR_NOT_EQUAL] = TRUE;
    } else {
        SdirWriteString(_T("Operator error in: "));
        return FALSE;
    }

    //
    //  Maybe the operator specified was valid but not supported
    //  by this type.
    //

    if (ThisApply->CompareFn == NULL) {
        SdirWriteString(_T("Operator not supported in: "));
        return FALSE;
    }

    //
    //  If we fail to capture this, ignore it and move on to the
    //  next entry, clobbering this structure again.
    //

    if (!SdirOptions[FeatureNumber].GenerateFromStringFn(&ThisApply->CompareEntry, Criteria)) {
        SdirWriteString(_T("Could not parse evalutation criteria in: "));
        return FALSE;
    }

    //
    //  Mark anything we're filtering by for collection
    //

    Feature = SdirFeatureByOptionNumber(FeatureNumber);
    Feature->Flags |= SDIR_FEATURE_COLLECT;
    return TRUE;
}



PSDIR_ATTRIBUTE_APPLY SdirAttributeApply = NULL; 
const
CHAR SdirAttributeDefaultApplyString[] = 
    "fa&r,magenta;"
    "fa&D,lightmagenta;"
    "fa&R,green;"
    "fa&H,green;"
    "fa&S,green;"
    "fe=bat,lightred;"
    "fe=cmd,lightred;"
    "fe=com,lightcyan;"
    "fe=dll,cyan;"
    "fe=doc,white;"
    "fe=docx,white;"
    "fe=exe,lightcyan;"
    "fe=htm,white;"
    "fe=html,white;"
    "fe=pdf,white;"
    "fe=pl,red;"
    "fe=ppt,white;"
    "fe=pptx,white;"
    "fe=ps1,lightred;"
    "fe=psd1,red;"
    "fe=psm1,red;"
    "fe=sys,cyan;"
    "fe=xls,white;"
    "fe=xlsx,white";

BOOL
SdirParseAttributeApplyString()
{
    TCHAR SingleSwitch[20];
    LPTSTR Next;
    LPTSTR This;
    LPTSTR Operator;
    LPTSTR Criteria;
    LPTSTR Attribute;
    ULONG i;
    PSDIR_ATTRIBUTE_APPLY ThisApply;
    LPTSTR SdirAttributeCombinedApplyString = NULL;
    LPTSTR TokContext = NULL;
    DWORD  SdirColorPrependLength = 0;
    DWORD  SdirColorAppendLength  = 0;
    DWORD  SdirColorReplaceLength = 0;
    DWORD  SdirColorCustomLength = 0;

    //
    //  Load any user specified colors from the environment.  Prepend values go before
    //  any default; replace values supersede any default; and append values go last.
    //  We need to insert semicolons between these values if they're specified.
    //
    //  First, count how big the allocation needs to be and allocate it.
    //

#ifdef SDIR_SUPPORT_CUSTOM_FILECOLOR
    SdirColorPrependLength = GetEnvironmentVariable(_T("SDIR_COLOR_PREPEND"), NULL, 0);
    SdirColorReplaceLength = GetEnvironmentVariable(_T("SDIR_COLOR_REPLACE"), NULL, 0);
    SdirColorAppendLength  = GetEnvironmentVariable(_T("SDIR_COLOR_APPEND"), NULL, 0);
    if (Opts->CustomFileColor != NULL) {
        SdirColorCustomLength = Opts->CustomFileColorLength;
    }
#endif

    SdirAttributeCombinedApplyString = HeapAlloc(GetProcessHeap(), 0, (SdirColorPrependLength + 1 + SdirColorAppendLength + 1 + SdirColorReplaceLength + 1 + SdirColorCustomLength + sizeof(SdirAttributeDefaultApplyString)) * sizeof(TCHAR));

    if (SdirAttributeCombinedApplyString == NULL) {
        return FALSE;
    }

    //
    //  Now, load any environment variables into the buffer.  If replace isn't specified,
    //  we use the hardcoded default.
    //

    i = 0;
#if SDIR_SUPPORT_CUSTOM_FILECOLOR
    if (SdirColorPrependLength) {
        GetEnvironmentVariable(_T("SDIR_COLOR_PREPEND"), SdirAttributeCombinedApplyString, SdirColorPrependLength);
        i += SdirColorPrependLength;
        SdirAttributeCombinedApplyString[i - 1] = ';';
    }

    if (SdirColorCustomLength) {
        _stprintf_s(SdirAttributeCombinedApplyString + i, SdirColorCustomLength + 1, _T("%s"), Opts->CustomFileColor);
        i += SdirColorCustomLength;
        SdirAttributeCombinedApplyString[i] = ';';
        i++;
    }

    if (SdirColorReplaceLength) {
        GetEnvironmentVariable(_T("SDIR_COLOR_REPLACE"), SdirAttributeCombinedApplyString + i, SdirColorReplaceLength);
        i += SdirColorReplaceLength - 1;
    } else {
#endif
        _stprintf_s(SdirAttributeCombinedApplyString + i, 
                    sizeof(SdirAttributeDefaultApplyString) / sizeof(SdirAttributeDefaultApplyString[0]),
                    _T("%hs"),
                    SdirAttributeDefaultApplyString);
        i += sizeof(SdirAttributeDefaultApplyString) / sizeof(SdirAttributeDefaultApplyString[0]) - 1;
#if SDIR_SUPPORT_CUSTOM_FILECOLOR
    }

    if (SdirColorAppendLength) {
        SdirAttributeCombinedApplyString[i] = ';';
        i += 1;
        GetEnvironmentVariable(_T("SDIR_COLOR_APPEND"), SdirAttributeCombinedApplyString + i, SdirColorAppendLength);
    }
#endif

    //
    //  First count the number of rules so we can allocate the correct amount
    //  of memory for the set.  For ease of use, we allocate an additional one
    //  as a "null terminator."
    //

    This = SdirAttributeCombinedApplyString;
    i = 1;

    while (This != NULL) {

        i++;
        Next = _tcschr(This, ';');

        if (Next) {
            This = Next + 1;
        } else {
            This = NULL;
        }
    }

    if (Opts->CustomFileFilter != NULL) {
        This = Opts->CustomFileFilter;

        while (This != NULL) {

            i++;
            Next = _tcschr(This, ';');

            if (Next) {
                This = Next + 1;
            } else {
                This = NULL;
            }
        }
    }

    SdirAttributeApply = HeapAlloc(GetProcessHeap(), 0, i * sizeof(SDIR_ATTRIBUTE_APPLY));
    if (SdirAttributeApply == NULL) {
        HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
        return FALSE;
    }

    ZeroMemory(SdirAttributeApply, i * sizeof(SDIR_ATTRIBUTE_APPLY));
    ThisApply = SdirAttributeApply;

    //
    //  First, process any filter/exclusion list, if present.
    //

    if (Opts->CustomFileFilter != NULL) {
        This = _tcstok_s(Opts->CustomFileFilter, _T(";"), &TokContext);

        //
        //  Now go through the rules again and populate the memory we just
        //  allocated with the comparison functions, criteria and attributes.
        //
    
        while (This != NULL) {
    
            //
            //  Ignore any leading spaces
            //
    
            while(*This == ' ') {
                This++;
            }
    
            //
            //  Count the number of chars to the operator.  Copy that
            //  many into the switch buffer, but not more than the
            //  size of the buffer.  We increment i to ensure we have
            //  space for the null terminator as well as the
            //  substring we want to process.
            //
    
            i = (ULONG)_tcscspn(This, _T("&<>=!"));
            if (i >= sizeof(SingleSwitch)/sizeof(SingleSwitch[0])) {
                i = sizeof(SingleSwitch)/sizeof(SingleSwitch[0]);
            } else {
                i++;
            }
    
            //
            //  Here we just give up on the 'secure' CRT things.  We
            //  want to copy the switch, up to the size of the buffer,
            //  and we're doing it manually to avoid puking because
            //  the source string may still (will still) have characters
            //  beyond this number.
            //
    
            memcpy(SingleSwitch, This, i * sizeof(TCHAR));
            SingleSwitch[i - 1] = '\0';
    
            Operator = This + i - 1;
    
            //
            //  Count the length of the operator and verify it's
            //  sane.
            //
    
            i = (ULONG)_tcsspn(Operator, _T("&<>=!"));
    
            if (i != 1 && i != 2) {
                SdirWriteString(_T("Operator error in: "));
                goto error_return;
            }
    
            //
            //  Count the length of the criteria and null terminate
            //  it for convenience.
            //
    
            Criteria = Operator + i;
    
            for (i = 0; i < SdirGetNumSdirOptions(); i++) {
                if (_tcsicmp(SingleSwitch, SdirOptions[i].Switch) == 0 &&
                    SdirOptions[i].GenerateFromStringFn != NULL) {
    
                    if (!SdirPopulateApplyEntry(ThisApply, Operator, Criteria, i)) {
                        goto error_return;
                    }
    
                    ThisApply->Attribute.Ctrl = SDIR_ATTRCTRL_HIDE;
                    ThisApply->Attribute.Win32Attr = 0;
    
                    ThisApply++;
                    break;
                }
            }
    
            if (i == SdirGetNumSdirOptions()) {
                SdirWriteString(_T("Criteria error in: "));
                goto error_return;
            }
    
            This = _tcstok_s(NULL, _T(";"), &TokContext);
        }

        //
        //  Since _tcstok just turned this into swiss cheese, deallocate it.
        //

        HeapFree(GetProcessHeap(), 0, Opts->CustomFileFilter);
        Opts->CustomFileFilter = NULL;
    }

    This = _tcstok_s(SdirAttributeCombinedApplyString, _T(";"), &TokContext);

    //
    //  Now go through the rules again and populate the memory we just
    //  allocated with the comparison functions, criteria and attributes.
    //

    while (This != NULL) {

        //
        //  Ignore any leading spaces
        //

        while(*This == ' ') {
            This++;
        }

        //
        //  Count the number of chars to the operator.  Copy that
        //  many into the switch buffer, but not more than the
        //  size of the buffer.  We increment i to ensure we have
        //  space for the null terminator as well as the
        //  substring we want to process.
        //

        i = (ULONG)_tcscspn(This, _T("&<>=!"));
        if (i >= sizeof(SingleSwitch)/sizeof(SingleSwitch[0])) {
            i = sizeof(SingleSwitch)/sizeof(SingleSwitch[0]);
        } else {
            i++;
        }

        //
        //  Here we just give up on the 'secure' CRT things.  We
        //  want to copy the switch, up to the size of the buffer,
        //  and we're doing it manually to avoid puking because
        //  the source string may still (will still) have characters
        //  beyond this number.
        //

        memcpy(SingleSwitch, This, i * sizeof(TCHAR));
        SingleSwitch[i - 1] = '\0';

        Operator = This + i - 1;

        //
        //  Count the length of the operator and verify it's
        //  sane.
        //

        i = (ULONG)_tcsspn(Operator, _T("&<>=!"));

        if (i != 1 && i != 2) {
            SdirWriteString(_T("Operator error in: "));
            goto error_return;
        }

        //
        //  Count the length of the criteria and null terminate
        //  it for convenience.
        //

        Criteria = Operator + i;

        i = (ULONG)_tcscspn(Criteria, _T(","));
        if (Criteria[i] == '\0') {
            SdirWriteString(_T("Attribute error in: "));
            goto error_return;
        }

        Criteria[i] = '\0';
        Attribute = Criteria + i + 1;

        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(SingleSwitch, SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].GenerateFromStringFn != NULL) {

                if (!SdirPopulateApplyEntry(ThisApply, Operator, Criteria, i)) {
                    goto error_return;
                }

                ThisApply->Attribute = SdirAttributeFromString(Attribute);
                ThisApply->Attribute.Ctrl &= ~(SDIR_ATTRCTRL_INVALID_FILE);

                ThisApply++;
                break;
            }
        }

        if (i == SdirGetNumSdirOptions()) {
            SdirWriteString(_T("Criteria error in: "));
            goto error_return;
        }

        This = _tcstok_s(NULL, _T(";"), &TokContext);
    }

    HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
    return TRUE;

error_return:

    SdirWriteString(This);
    SdirWriteString(_T("\n"));

    HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
    return FALSE;
}

BOOL
SdirParseMetadataAttributeString()
{
#ifdef SDIR_SUPPORT_CUSTOM_METACOLOR
    TCHAR SingleElement[256];
    TCHAR SingleSwitch[20];
    LPTSTR Next;
    LPTSTR This;
    LPTSTR Attribute;
    ULONG i;
    LPTSTR SdirApplyString = NULL;
    DWORD  SdirColorMetadataLength = 0;
    PSDIR_FEATURE Feature;

    //
    //  Load any user specified colors from the environment.
    //
    //  First, count how big the allocation needs to be and allocate it.
    //

    SdirColorMetadataLength = GetEnvironmentVariable(_T("SDIR_COLOR_METADATA"), NULL, 0);
    if (SdirColorMetadataLength == 0) {
        return TRUE;
    }
    SdirApplyString = HeapAlloc(GetProcessHeap(), 0, (SdirColorMetadataLength + 1) * sizeof(TCHAR));

    if (SdirApplyString == NULL) {
        return FALSE;
    }

    //
    //  Now, load any environment variables into the buffer.
    //

    GetEnvironmentVariable(_T("SDIR_COLOR_METADATA"), SdirApplyString, SdirColorMetadataLength);

    This = SdirApplyString;

    //
    //  Now go through the rules again and populate the memory we just
    //  allocated with the comparison functions, criteria and attributes.
    //

    while (This != NULL) {

        //
        //  We have a mutable buffer since we just allocated it.
        //  NULL-ing as we go is really a hint to new 'secure'
        //  versions of the CRT that we don't intend to copy more
        //  TCHARs than this, so it doesn't need to gracelessly
        //  puke.
        //

        Next = _tcschr(This, ';');
        if (Next) {
            *Next = '\0';
        }

        //
        //  Ignore any leading spaces
        //

        while(*This == ' ') {
            This++;
        }

        //
        //  We want to copy as many tchars as there are between two
        //  elements, but not more than the buffer size.  Note that
        //  the printf routine below will do the right thing at end
        //  of string so it's okay to pass a larger buffer size
        //  there.
        //

        i = sizeof(SingleElement)/sizeof(SingleElement[0]);

        if (Next && ((ULONG)(Next - This) < i)) {
            i = (ULONG)(Next - This + 1);
        }

        //
        //  If this element is empty, ignore it
        //

        if (i == 0 || *This == '\0') {
            goto next_element;
        }

        _stprintf_s(SingleElement, i, _T("%s"), This);
        SingleElement[i - 1] = '\0';

        //
        //  Count the number of chars to the operator.  Copy that
        //  many into the switch buffer, but not more than the
        //  size of the buffer.  We increment i to ensure we have
        //  space for the null terminator as well as the
        //  substring we want to process.
        //

        i = (ULONG)_tcscspn(SingleElement, _T(","));
        if (i >= sizeof(SingleSwitch)/sizeof(SingleSwitch[0])) {
            i = sizeof(SingleSwitch)/sizeof(SingleSwitch[0]);
        } else {
            i++;
        }

        //
        //  Here we just give up on the 'secure' CRT things.  We
        //  want to copy the switch, up to the size of the buffer,
        //  and we're doing it manually to avoid puking because
        //  the source string may still (will still) have characters
        //  beyond this number.
        //

        memcpy(SingleSwitch, SingleElement, i * sizeof(TCHAR));
        SingleSwitch[i - 1] = '\0';

        Attribute = SingleElement + i;

        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(SingleSwitch, SdirOptions[i].Switch) == 0) {

                SDIR_COLOR_ATTRIBUTES HighlightColor;
                Feature = SdirFeatureByOptionNumber(i);

                if (Feature->Flags & SDIR_FEATURE_FIXED_COLOR) {
                    SdirWriteString(_T("Color cannot be changed for attribute in: "));
                    goto error_return;
                }

                //
                //  Set this piece of metadata to use the desired attributes.
                //  If it's invalid, just ignore it.  Hack around formats so
                //  we can remember to ignore this color and use the file's
                //  color in the flags field and retain color in the attribute.
                //

                HighlightColor = SdirAttributeFromString(Attribute);
                if (HighlightColor.Ctrl & SDIR_ATTRCTRL_INVALID_METADATA) {
                    SdirWriteString(_T("Invalid color specified in: "));
                    goto error_return;
                    break;
                }

                Feature->HighlightColor = SdirResolveWindowColorComponents(HighlightColor, FALSE);

                if (HighlightColor.Ctrl & SDIR_ATTRCTRL_FILE) {
                    Feature->Flags |= SDIR_FEATURE_USE_FILE_COLOR;
                } else {
                    Feature->Flags &= ~(SDIR_FEATURE_USE_FILE_COLOR);
                }
                break;
            }
        }

next_element:

        if (Next) {
            This = Next + 1;
        } else {
            This = NULL;
        }
    }

    HeapFree(GetProcessHeap(), 0, SdirApplyString);
    return TRUE;

error_return:

    SdirWriteString(SingleElement);
    SdirWriteString(_T("\n"));

    HeapFree(GetProcessHeap(), 0, SdirApplyString);
    return FALSE;
#else
    return TRUE;
#endif
}

BOOL
SdirApplyAttribute(
    __in PSDIR_DIRENT DirEnt,
    __out PSDIR_COLOR_ATTRIBUTES Attribute
    )
{
    PSDIR_ATTRIBUTE_APPLY ThisApply;
    SDIR_COLOR_ATTRIBUTES ThisAttribute = {SDIR_ATTRCTRL_WINDOW_BG | SDIR_ATTRCTRL_WINDOW_FG, 0};

    if (SdirAttributeApply == NULL) {
        return FALSE;
    }
    
    ThisApply = SdirAttributeApply;

    while (ThisApply->CompareFn) {

        if (ThisApply->TruthStates[ThisApply->CompareFn(DirEnt, &ThisApply->CompareEntry)]) {
            ThisAttribute = SdirCombineColors(ThisAttribute, ThisApply->Attribute);
            if ((ThisAttribute.Ctrl & SDIR_ATTRCTRL_CONTINUE) == 0) {

                ThisAttribute = SdirResolveWindowColorComponents(ThisAttribute, FALSE);

                if (ThisAttribute.Ctrl & SDIR_ATTRCTRL_INVERT) {
                    ThisAttribute.Win32Attr = (UCHAR)(((ThisAttribute.Win32Attr & 0x0F) << 4) + ((ThisAttribute.Win32Attr & 0xF0) >> 4));
                }

                *Attribute = ThisAttribute;
                return TRUE;
            }

            ThisAttribute.Ctrl = ThisAttribute.Ctrl & ~(SDIR_ATTRCTRL_CONTINUE);
        }

        ThisApply++;
    }

    //
    //  We do let the user explicitly request black on black, but
    //  if we ended the search due to unbounded continues, return
    //  what we have.
    //

    if (ThisAttribute.Ctrl & SDIR_ATTRCTRL_TERMINATE_MASK || ThisAttribute.Win32Attr != 0) {

        *Attribute = ThisAttribute;
        return TRUE;
    }

    *Attribute = SdirDefaultColor;
    
    return FALSE;
}

BOOL
SdirSetColorToWin32(
    __out PSDIR_COLOR_ATTRIBUTES Attributes,
    __in UCHAR Win32Attribute
    )
{
    Attributes->Ctrl = 0;
    Attributes->Win32Attr = Win32Attribute;
    return TRUE;
}

SDIR_COLOR_ATTRIBUTES
SdirCombineColors(
    __in SDIR_COLOR_ATTRIBUTES Color1,
    __in SDIR_COLOR_ATTRIBUTES Color2
    )
{
    SDIR_COLOR_ATTRIBUTES Result;

    Result.Ctrl = 0;
    Result.Win32Attr = 0;

    Result.Ctrl = (Color1.Ctrl | Color2.Ctrl) & ~(SDIR_ATTRCTRL_WINDOW_BG|SDIR_ATTRCTRL_WINDOW_FG);

    //
    //  Treat the default window color as recessive.  If anything explicitly
    //  includes a color, that takes precedence.
    //

    if ((Color1.Ctrl & SDIR_ATTRCTRL_WINDOW_BG) != 0 &&
        (Color2.Ctrl & SDIR_ATTRCTRL_WINDOW_BG) != 0) {

        Result.Ctrl |= SDIR_ATTRCTRL_WINDOW_BG;
    } else {
        Result.Win32Attr |= (Color1.Win32Attr ^ Color2.Win32Attr) & 0xF0;
    }

    if ((Color1.Ctrl & SDIR_ATTRCTRL_WINDOW_FG) != 0 &&
        (Color2.Ctrl & SDIR_ATTRCTRL_WINDOW_FG) != 0) {

        Result.Ctrl |= SDIR_ATTRCTRL_WINDOW_FG;
    } else {
        Result.Win32Attr |= (Color1.Win32Attr ^ Color2.Win32Attr) & 0x0F;
    }

    return Result;
}

SDIR_COLOR_ATTRIBUTES
SdirResolveWindowColorComponents(
    __in SDIR_COLOR_ATTRIBUTES Color,
    __in BOOL RetainWindowCtrlFlags
    )
{
    SDIR_COLOR_ATTRIBUTES NewColor = Color;

    if (NewColor.Ctrl & SDIR_ATTRCTRL_WINDOW_BG) {
        NewColor.Win32Attr = (UCHAR)(Opts->PreviousAttributes.Win32Attr & 0xF0) + (NewColor.Win32Attr & 0xF);
    }

    if (NewColor.Ctrl & SDIR_ATTRCTRL_WINDOW_FG) {
        NewColor.Win32Attr = (UCHAR)(NewColor.Win32Attr & 0xF0) + (Opts->PreviousAttributes.Win32Attr & 0xF);
    }

    if (!RetainWindowCtrlFlags) {
        NewColor.Ctrl &= ~(SDIR_ATTRCTRL_WINDOW_BG|SDIR_ATTRCTRL_WINDOW_FG);
    }

    return NewColor;
}

BOOL
SdirAreColorsIdentical(
    __in SDIR_COLOR_ATTRIBUTES Color1,
    __in SDIR_COLOR_ATTRIBUTES Color2
    )
{
    if (Color1.Win32Attr == Color2.Win32Attr) {
        return TRUE;
    }
    return FALSE;
}

// vim:sw=4:ts=4:et:
