/*
 * INIT.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements initialization support including argument parsing
 * and initializing default options.
 *
 * Copyright (c) 2014-2017 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"
#include "msupd.h"

BOOL WINAPI
SdirRestoreColor (
    __in DWORD dwHandlerType
    )
{
    UNREFERENCED_PARAMETER(dwHandlerType);
    Opts->Cancelled = TRUE;
    return TRUE;
}

BOOL
SdirAppInitialize()
{
    CONSOLE_SCREEN_BUFFER_INFO ScreenInfo;
    HANDLE hConsoleOutput = GetStdHandle(STD_OUTPUT_HANDLE);
    HANDLE hKernel32 = GetModuleHandle(_T("KERNEL32"));
    DISABLE_WOW_REDIRECT_FN DisableWow = (DISABLE_WOW_REDIRECT_FN)GetProcAddress(hKernel32, "Wow64DisableWow64FsRedirection");
    HANDLE ProcessToken;
    DWORD CurrentMode;

    Opts = HeapAlloc(GetProcessHeap(), 0, sizeof(SDIR_OPTS));
    if (Opts == NULL) {
        return FALSE;
    }
    ZeroMemory(Opts, sizeof(SDIR_OPTS));

    Summary = HeapAlloc(GetProcessHeap(), 0, sizeof(SDIR_SUMMARY));
    if (Summary == NULL) {
        return FALSE;
    }
    ZeroMemory(Summary, sizeof(SDIR_SUMMARY));

    //
    //  For simplicity, initialize this now.  On failure we restore to
    //  this value.  Hopefully we'll find the correct value before any
    //  failure can occur.
    //

    Opts->PreviousAttributes = SdirDefaultColor;

    Opts->PathPrefixChar = '?';

    //
    //  Check if we're talking to a console or some other kind of object
    //  (file, pipe, etc.)
    //

    if (!GetConsoleMode(hConsoleOutput, &CurrentMode)) {
        Opts->OutputWithConsoleApi = FALSE;
        SdirQueryVtSupport();
        Opts->NewLine = _T("\r\n");
    } else {
        Opts->OutputHasAutoLineWrap = TRUE;
        Opts->OutputExtendedCharacters = TRUE;
        Opts->OutputWithConsoleApi = TRUE;
        Opts->NewLine = _T("\n");
        Opts->EnablePause = TRUE;
    }

    //
    //  Try to determine the console width so we can size columns.  We
    //  try to use the window size, not the buffer size.
    //  If we're not talking to a console, this might fail, and we have
    //  to fall back to some kind of default.
    //

    if (!SdirGetConsoleScreenBufferInfo(hConsoleOutput, &ScreenInfo)) {
        if (GetLastError() == ERROR_INVALID_HANDLE) {
            if (Opts->ConsoleWidth == 0) {
                Opts->ConsoleWidth = 100;
            }
            Opts->ConsoleBufferWidth = Opts->ConsoleWidth;
            Opts->EnablePause = FALSE;
        } else {
            SdirDisplayError(GetLastError(), _T("GetConsoleScreenBufferInfo"));
            return FALSE;
        }
    } else {
        Opts->ConsoleBufferWidth = 
        Opts->ConsoleWidth = ScreenInfo.dwSize.X;
        
        if (Opts->ConsoleWidth > (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1)) {
            Opts->ConsoleWidth = (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1);
        }
    
        Opts->ConsoleHeight = ScreenInfo.dwSize.Y;
        if (Opts->ConsoleHeight > (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1)) {
            Opts->ConsoleHeight = (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1);
        }

        //
        //  This is another kludge.  To make progress we need to be able
        //  to display one line of output between "press any key" prompts,
        //  so we need to assume the console can hold two lines.  If we
        //  don't make this assumption, we'll never go forward.
        //

        if (Opts->ConsoleHeight < 2) {
            Opts->ConsoleHeight = 2;
        }

        SdirSetColorToWin32(&Opts->PreviousAttributes, (UCHAR)(ScreenInfo.wAttributes & SDIR_ATTRIBUTE_FULLCOLOR_MASK));
    }

    SdirDefaultColor = SdirResolveWindowColorComponents(SdirDefaultColor, TRUE);

    if (Opts->ConsoleWidth > SDIR_MAX_WIDTH) {
        Opts->ConsoleWidth = SDIR_MAX_WIDTH;
    }

    //
    //  When running on WOW64, we don't want file system redirection,
    //  because we want people to be able to enumerate those paths
    //

    if (DisableWow) {
        PVOID DontCare;
        DisableWow(&DontCare);
    }

    //
    //  Attempt to enable backup privilege.  This allows us to enumerate and recurse
    //  through objects which normally ACLs would prevent.
    //

    if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &ProcessToken)) {
        struct {
            TOKEN_PRIVILEGES TokenPrivileges;
            LUID_AND_ATTRIBUTES BackupPrivilege;
        } PrivilegesToChange;

        LookupPrivilegeValue(NULL, SE_BACKUP_NAME, &PrivilegesToChange.TokenPrivileges.Privileges[0].Luid);

        PrivilegesToChange.TokenPrivileges.PrivilegeCount = 1;
        PrivilegesToChange.TokenPrivileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

        AdjustTokenPrivileges(ProcessToken, FALSE, (PTOKEN_PRIVILEGES)&PrivilegesToChange, sizeof(PrivilegesToChange), NULL, NULL);
        CloseHandle(ProcessToken);
    }

    //
    //  Try to locate FindFirstStream functions if they exist
    //

    Opts->FindFirstStreamW = (FIND_FIRST_STREAM_FN)GetProcAddress(hKernel32, "FindFirstStreamW");
    Opts->FindNextStreamW = (FIND_NEXT_STREAM_FN)GetProcAddress(hKernel32, "FindNextStreamW");
#ifdef UNICODE
    Opts->GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeW");
    Opts->GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExW");
#else
    Opts->GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeA");
    Opts->GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExA");
#endif
    Opts->GetFileInformationByHandleEx = (GET_FILE_INFORMATION_BY_HANDLE_EX_FN)GetProcAddress(hKernel32, "GetFileInformationByHandleEx");

    //
    //  Grab the version of the running OS so we can highlight binaries that
    //  need a newer one
    //

    Opts->OsVersion = GetVersion();
    if (LOBYTE(LOWORD(Opts->OsVersion)) < 4) {
        Opts->PathPrefixChar = '.';
    }

    //
    //  Look for Ctrl+C and restore color if we get one.
    //  Not much we can do on failure.
    //

    SetConsoleCtrlHandler(SdirRestoreColor, TRUE);

    return TRUE;
}

BOOL
SdirLoadVerFunctions()
{
    HANDLE hVerDll;

    hVerDll = LoadLibrary(_T("VERSION.DLL"));
    if (hVerDll == NULL) {
        return FALSE;
    }

#ifdef UNICODE
    Opts->GetFileVersionInfoSize = (GET_FILE_VERSION_INFO_SIZE_FN)GetProcAddress(hVerDll, "GetFileVersionInfoSizeW");
    Opts->GetFileVersionInfo = (GET_FILE_VERSION_INFO_FN)GetProcAddress(hVerDll, "GetFileVersionInfoW");
    Opts->VerQueryValue = (VER_QUERY_VALUE_FN)GetProcAddress(hVerDll, "VerQueryValueW");
#else
    Opts->GetFileVersionInfoSize = (GET_FILE_VERSION_INFO_SIZE_FN)GetProcAddress(hVerDll, "GetFileVersionInfoSizeA");
    Opts->GetFileVersionInfo = (GET_FILE_VERSION_INFO_FN)GetProcAddress(hVerDll, "GetFileVersionInfoA");
    Opts->VerQueryValue = (VER_QUERY_VALUE_FN)GetProcAddress(hVerDll, "VerQueryValueA");
#endif

    //
    //  If we failed to load all functions, close up and say we couldn't
    //  load any.
    //

    if (Opts->GetFileVersionInfoSize == NULL ||
        Opts->GetFileVersionInfo == NULL ||
        Opts->VerQueryValue == NULL) {

        FreeLibrary(hVerDll);
        Opts->GetFileVersionInfoSize = NULL;
        Opts->GetFileVersionInfo = NULL;
        Opts->VerQueryValue = NULL;
        return FALSE;
    }

    return TRUE;
}

BOOL
SdirAssembleUrl(
    LPTSTR szUrl,
    DWORD szUrlLength,
    LPTSTR szVerString,
    LPTSTR szFileName,
    LPTSTR szFileExtension
    )
{
    TCHAR szTempFileName[128];

    if (szFileName == NULL) {
        _stprintf_s(szTempFileName,
                    sizeof(szTempFileName)/sizeof(szTempFileName[0]),
                    _T("%s%s/sdir"),
#ifdef _M_AMD64
                    _T("amd64"),
#elif defined(_M_IX86)
                    _T("win32"),
#endif
#ifdef UNICODE
                    _T("-unicode")
#else
                    _T("-ansi")
#endif
            );

        szFileName = szTempFileName;
    }

    _stprintf_s(szUrl,
                szUrlLength,
                _T("http://www.malsmith.net/download/?obj=sdir/%s/%s%s"),
                szVerString?szVerString:
#if SDIR_BUILD_ID
                _T("latest-daily"),
#else
                _T("latest-stable"),
#endif
                szFileName,
                szFileExtension?szFileExtension:_T(".exe")
            );

    return TRUE;
}

typedef enum {
    SdirUpdateTypeBinary = 0,
    SdirUpdateTypeSource,
    SdirUpdateTypeSymbol,
    SdirUpdateTypeStable,
    SdirUpdateTypeDaily
} SDIR_UPDATE_TYPE;

#ifdef SDIR_SUPPORT_UPDATE
BOOL
SdirUpdateBinary(
    SDIR_UPDATE_TYPE UpdateType
    )
{
    TCHAR szAgent[128];
    TCHAR szUrl[250];
    TCHAR szVer[32];
    LPTSTR TargetName = NULL;
    SDIR_COLOR_ATTRIBUTES Color = SdirDefaultColor;
    UpdError Error = 0;

#if SDIR_BUILD_ID
    _stprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), _T("%i.%i.%i"), SDIR_VER_MAJOR, SDIR_VER_MINOR, SDIR_BUILD_ID);
#else
    _stprintf_s(szVer, sizeof(szVer)/sizeof(szVer[0]), _T("%i.%i"), SDIR_VER_MAJOR, SDIR_VER_MINOR);
#endif

    _stprintf_s(szAgent, sizeof(szAgent)/sizeof(szAgent[0]), _T("sdir %s"), szVer);

    if (UpdateType == SdirUpdateTypeBinary) {
        SdirAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), Opts->VersionToUpdateTo, NULL, NULL);
    } else if (UpdateType == SdirUpdateTypeSource) {
        SdirAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, _T("sdir-source"), _T(".zip"));
        TargetName = _T("sdir-source.zip");
    } else if (UpdateType == SdirUpdateTypeSymbol) {
        SdirAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), szVer, NULL, _T(".pdb"));
        TargetName = _T("sdir.pdb");
    } else if (UpdateType == SdirUpdateTypeStable) {
        SdirAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), _T("latest-stable"), NULL, NULL);
    } else if (UpdateType == SdirUpdateTypeDaily) {
        SdirAssembleUrl(szUrl, sizeof(szUrl)/sizeof(szUrl[0]), _T("latest-daily"), NULL, NULL);
    }
    SdirWriteString(_T("Updating from "));
    SdirWriteString(szUrl);
    SdirWriteString(_T("\n"));
    Error = UpdateBinaryFromUrl(szUrl, TargetName, szAgent);

    _stprintf_s(szAgent, sizeof(szAgent)/sizeof(szAgent[0]), _T("Update result: %s\n"), UpdateErrorString(Error));
    if (Error != UpdErrorSuccess) {
        Color = Opts->FtError.HighlightColor;
    }
    SdirWriteStringWithAttribute(szAgent, Color);
    return TRUE;
}
#endif

BOOL
SdirOptInitialize()
{
    ULONG i, j;

    //
    //  Calculate the amount of metadata in each column
    //

    Opts->MetadataWidth = 1;  // Column seperator

    //
    //  If no sorting algorithm was used, use the default that
    //  we prepopulated.
    //

    if (Opts->CurrentSort == 0) {
        Opts->CurrentSort++;
    }

    for (i = 0; i < SdirGetNumSdirOptions(); i++) {

        PSDIR_FEATURE Feature;

        Feature = SdirFeatureByOptionNumber(i);
       
        //
        //  If we're displaying or sorting, we need the data to use
        //

        if (Feature->Flags & SDIR_FEATURE_DISPLAY) {

            Feature->Flags |= SDIR_FEATURE_COLLECT;
        }

        for (j = 0; j < Opts->CurrentSort; j++) {

            if (SdirOptions[i].CompareFn == Opts->Sort[j].CompareFn) {
                Feature->Flags |= SDIR_FEATURE_COLLECT;
            }
        }

        //
        //  If we're displaying, we need space to display in
        //  

        if ((Feature->Flags & SDIR_FEATURE_DISPLAY) &&
               SdirOptions[i].WidthFn) {

            SDIR_COLOR_ATTRIBUTES Attr = {0};

            Opts->MetadataWidth += SdirOptions[i].WidthFn(NULL, Attr, NULL);
        }
    }

    //
    //  If we need to count the average link size, we need to know the link
    //  count too
    //

    if (Opts->EnableAverageLinkSize) {
        Opts->FtLinkCount.Flags |= SDIR_FEATURE_COLLECT;
    }

    return TRUE;
}

#ifdef SDIR_SUPPORT_RECURSIVE
BOOL
SdirSetSubdirWalk(
    __in LPTSTR SubDirPath
    )
{
    if (!SdirGetFullPathNameReturnAllocation(SubDirPath, &Opts->SubDirWalk, NULL)) {
        SdirDisplayError(GetLastError(), _T("SdirGetFullPathName"));
        return FALSE;
    }

    return TRUE;
}
#endif


BOOL
SdirParseOpt (
    __in LPTSTR Opt
    )
{
    BOOL OptParsed = FALSE;
    DWORD i, j;

    if (Opt[0] == 'a') {
#ifdef SDIR_SUPPORT_AVERAGE_LINK_SIZE
        if (Opt[1] == 'l') {
            if (Opt[2] == 'n') {
                Opts->EnableAverageLinkSize = FALSE;
                OptParsed = TRUE;
            } else if (Opt[2] == '\0') {
                Opts->EnableAverageLinkSize = TRUE;
                OptParsed = TRUE;
            }
        }
#endif
    } else if (Opt[0] == 'b') {
#ifdef SDIR_SUPPORT_RECURSIVE
        if (Opt[1] == 'r') {
            Opts->BriefRecurseDepth = SdirStringToNum32(&Opt[2], NULL);
            if (Opts->BriefRecurseDepth == 0) {
                Opts->BriefRecurseDepth = UINT_MAX;
            }
            OptParsed = TRUE;

            if (Opts->SubDirWalk == NULL) {
                if (!SdirSetSubdirWalk(_T("."))) {
                    return FALSE;
                }
            }
        } else if (Opt[1] == 's') {
            LARGE_INTEGER FileSize;
            if (Opts->BriefRecurseDepth == 0) {
                Opts->BriefRecurseDepth = UINT_MAX;
            }
            FileSize = SdirStringToFileSize(&Opt[2]);
            Opts->BriefRecurseSize = SdirFileSizeFromLargeInt(&FileSize);
            OptParsed = TRUE;

            if (Opts->SubDirWalk == NULL) {
                if (!SdirSetSubdirWalk(_T("."))) {
                    return FALSE;
                }
            }
        }
#endif
    } else if (Opt[0] == 'c') {
        if (Opt[1] == 'w') {
            Opts->ConsoleWidth = SdirStringToNum32(&Opt[2], NULL);
            if (Opts->ConsoleWidth > SDIR_MAX_WIDTH) {
                Opts->ConsoleWidth = SDIR_MAX_WIDTH;
            }
            if (!Opts->OutputHasAutoLineWrap) {
                Opts->ConsoleBufferWidth = Opts->ConsoleWidth;
            }
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 'd') {
        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].Default.Flags & SDIR_FEATURE_ALLOW_DISPLAY) {

                PSDIR_FEATURE Feature;

                Feature = SdirFeatureByOptionNumber(i);
                Feature->Flags |= SDIR_FEATURE_DISPLAY;
                OptParsed = TRUE;
                break;
            }
        }
    } else if (Opt[0] == 'h') {
        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].Default.Flags & SDIR_FEATURE_ALLOW_DISPLAY) {

                PSDIR_FEATURE Feature;

                Feature = SdirFeatureByOptionNumber(i);
                Feature->Flags &= ~(SDIR_FEATURE_DISPLAY);
                OptParsed = TRUE;
                break;
            }
        }
    } else if (Opt[0] == 'f') {
        if (Opt[1] == 'c') {
            DWORD Len = (DWORD)_tcslen(&Opt[2]);
            if (Len > 0) {
                Opts->CustomFileColor = HeapAlloc(GetProcessHeap(), 0, (Len + 1) * sizeof(TCHAR));
                if (Opts->CustomFileColor == NULL) {
                    return FALSE;
                }
                _stprintf_s(Opts->CustomFileColor, Len + 1, _T("%s"), &Opt[2]);
                Opts->CustomFileColorLength = Len;
                OptParsed = TRUE;
            }
        } else if (Opt[1] == 'e') {
            DWORD Len = (DWORD)_tcslen(&Opt[2]);
            if (Len > 0) {
                Opts->CustomFileFilter = HeapAlloc(GetProcessHeap(), 0, (Len + 1) * sizeof(TCHAR));
                if (Opts->CustomFileFilter == NULL) {
                    return FALSE;
                }
                _stprintf_s(Opts->CustomFileFilter, Len + 1, _T("%s"), &Opt[2]);
                Opts->CustomFileFilterLength = Len;
                OptParsed = TRUE;
            }
        }
#ifdef SDIR_SUPPORT_UPDATE
    } else if (Opt[0] == 'g') {
        if (Opt[1] == 'e' && Opt[2] == 't' && Opt[3] == '-') {
            if (_tcscmp(&Opt[4], _T("daily")) == 0) {
                Opts->UpdateDaily = TRUE;
                OptParsed = TRUE;
            }
            if (_tcscmp(&Opt[4], _T("source")) == 0) {
                Opts->UpdateSource = TRUE;
                OptParsed = TRUE;
            }
            if (_tcscmp(&Opt[4], _T("stable")) == 0) {
                Opts->UpdateStable = TRUE;
                OptParsed = TRUE;
            }
            if (_tcscmp(&Opt[4], _T("symbol")) == 0) {
                Opts->UpdateSymbol = TRUE;
                OptParsed = TRUE;
            }
            if (_tcscmp(&Opt[4], _T("update")) == 0) {
                Opts->UpdateBinary = TRUE;
                OptParsed = TRUE;
            }
            if (Opt[4] == 'v' && Opt[5] == 'e' && Opt[6] == 'r' && Opt[7] == '-') {
                Opts->UpdateToVersion = TRUE;
                Opts->VersionToUpdateTo = &Opt[8];
                OptParsed = TRUE;
            }
        }
#endif
    } else if (Opt[0] == 'l') {
        if (Opt[1] == 'n') {
            Opts->TraverseLinks = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts->TraverseLinks = TRUE;
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 's' || Opt[0] == 'i') {

        //
        //  If we don't have space for it, don't even try
        //

        if (Opts->CurrentSort < sizeof(Opts->Sort)/sizeof(Opts->Sort[0])) {

            //
            //  Find which option it is
            //

            for (i = 0; i < SdirGetNumSdirOptions(); i++) {
                if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                    SdirOptions[i].Default.Flags & SDIR_FEATURE_ALLOW_SORT) {

                    //
                    //  See if it's been specified before and silently discard
                    //  it.
                    //

                    for (j = 0; j < Opts->CurrentSort; j++) {
                        if (Opts->Sort[j].CompareFn == SdirOptions[i].CompareFn) {
                            OptParsed = TRUE;
                            break;
                        }
                    }

                    if (OptParsed) {
                        break;
                    }

                    Opts->Sort[Opts->CurrentSort].CompareFn = SdirOptions[i].CompareFn;

                    if (Opt[0] == 's') {
                        Opts->Sort[Opts->CurrentSort].CompareBreakCondition = SDIR_GREATER_THAN;
                        Opts->Sort[Opts->CurrentSort].CompareInverseCondition = SDIR_LESS_THAN;
                    } else {
                        Opts->Sort[Opts->CurrentSort].CompareBreakCondition = SDIR_LESS_THAN;
                        Opts->Sort[Opts->CurrentSort].CompareInverseCondition = SDIR_GREATER_THAN;
                    }
                    Opts->CurrentSort++;
                    OptParsed = TRUE;

                    break;
                }
            }

        } else {
        }
    } else if (Opt[0] == 'p') {
        if (Opt[1] == 'n') {
            Opts->EnablePause = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            if (Opts->OutputWithConsoleApi) {
                Opts->EnablePause = TRUE;
            }
            OptParsed = TRUE;
        }
#ifdef SDIR_SUPPORT_RECURSIVE
    } else if (Opt[0] == 'r') {

        LPTSTR SubDirPath = &Opt[1];

        OptParsed = TRUE;
        if (SubDirPath[0] == '\0') {
            SubDirPath = _T(".");
        }

        if (!SdirSetSubdirWalk(SubDirPath)) {
            return FALSE;
        }
#endif

    } else if (Opt[0] == 't') {
        if (Opt[1] == 'n') {
            Opts->EnableNameTruncation = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts->EnableNameTruncation = TRUE;
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 'u') {
#ifdef UNICODE
        if (Opt[1] == 'n') {
            Opts->OutputIsUnicode = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts->OutputIsUnicode = TRUE;
            OptParsed = TRUE;
        }
#endif
    }

    return OptParsed;
}

BOOL
SdirParseArgs (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    ULONG  CurrentArg;
    BOOL   OptParsed;
    BOOL   DisplayUsage;

    TCHAR  EnvOpts[200];
    LPTSTR Opt;
    LPTSTR TokContext = NULL;
    DWORD  i;

    //
    //  Default to name sorting.  If something else is specified
    //  we clobber this entry.
    //

    Opts->CurrentSort = 0;
    Opts->Sort[0].CompareFn = SdirCompareFileName;
    Opts->Sort[0].CompareBreakCondition = SDIR_GREATER_THAN;
    Opts->Sort[0].CompareInverseCondition = SDIR_LESS_THAN;

    Opts->EnableNameTruncation = TRUE;

    Opts->SubDirWalk = NULL;

    //
    //  Default to ANSI files
    //

    if (!Opts->OutputWithConsoleApi) {
        Opts->OutputIsUnicode = FALSE;
    } else {
        Opts->OutputIsUnicode = TRUE;
    }

    for (i = 0; i < SdirGetNumSdirOptions(); i++) {
        PSDIR_FEATURE Feature;

        Feature = SdirFeatureByOptionNumber(i);
        Feature->Flags = SdirOptions[i].Default.Flags;
        Feature->HighlightColor = SdirResolveWindowColorComponents(SdirOptions[i].Default.HighlightColor, TRUE);
    }

    if (GetEnvironmentVariable(_T("SDIR_OPTS"), EnvOpts, sizeof(EnvOpts)/sizeof(EnvOpts[0]))) {
        EnvOpts[sizeof(EnvOpts)/sizeof(EnvOpts[0]) - 1] = '\0';
        Opt = _tcstok_s(EnvOpts, _T(" "), &TokContext);

        while (Opt) {

            if (Opt[0] == '-' || Opt[0] == '/') {

                Opt++;

                OptParsed = FALSE;

                OptParsed = SdirParseOpt(Opt);

                if (!OptParsed) {
                    SdirWriteString(_T("Unknown environment option: "));
                    SdirWriteString(Opt);
                    SdirWriteString(_T("\n"));
                    SdirUsage(argc, argv);
                    return FALSE;
                }
            }

            Opt = _tcstok_s(NULL, _T(" "), &TokContext);
        }
    }

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++) {

        if (argv[CurrentArg][0] == '-' || argv[CurrentArg][0] == '/') {
            Opt = &argv[CurrentArg][1];

            OptParsed = FALSE;
            DisplayUsage = FALSE;

            OptParsed = SdirParseOpt(Opt);

            if (_tcsicmp(Opt, _T("help")) == 0) {
                DisplayUsage = TRUE;
                OptParsed = TRUE;
            } else if (_tcsicmp(Opt, _T("?")) == 0) {
                DisplayUsage = TRUE;
                OptParsed = TRUE;
            } else if (_tcsicmp(Opt, _T("v")) == 0) {
                DisplayUsage = TRUE;
                OptParsed = TRUE;
            }

            if (DisplayUsage || !OptParsed) {
                if (!OptParsed) {
                    SdirWriteString(_T("Unknown argument: "));
                    SdirWriteString(argv[CurrentArg]);
                    SdirWriteString(_T("\n"));
                }
                SdirUsage(argc, argv);
                return FALSE;
            }
        }
    }

#ifdef SDIR_SUPPORT_UPDATE
    if (Opts->UpdateBinary || Opts->UpdateToVersion) {
        SdirUpdateBinary(SdirUpdateTypeBinary);
        return FALSE;
    }
    if (Opts->UpdateStable) {
        SdirUpdateBinary(SdirUpdateTypeStable);
        return FALSE;
    }
    if (Opts->UpdateDaily) {
        SdirUpdateBinary(SdirUpdateTypeDaily);
        return FALSE;
    }
    if (Opts->UpdateSource) {
        SdirUpdateBinary(SdirUpdateTypeSource);
        return FALSE;
    }
    if (Opts->UpdateSymbol) {
        SdirUpdateBinary(SdirUpdateTypeSymbol);
        return FALSE;
    }
#endif
    return TRUE;
}

BOOL
SdirInit(
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    if (!SdirAppInitialize()) {
        return FALSE;
    }

    if (!SdirParseArgs(argc, argv)) {
        return FALSE;
    }

    if (!SdirOptInitialize()) {
        return FALSE;
    }

    if (!SdirParseAttributeApplyString()) {
        return FALSE;
    }

    if (!SdirParseMetadataAttributeString()) {
        return FALSE;
    }

    return TRUE;
}

// vim:sw=4:ts=4:et:
